@@ -80,14 +80,14 @@ typedef struct pipeinfo_T {
 } pipeinfo_T;
 #define PIPEINFO_INIT { -1, { -1, -1 }, }
 
-/* values used to specify the way of command search. */
+/* values used to specify the behavior of command search. */
 enum srchcmdtype_T {
-    sct_external = 1 << 0,  /* search external commands */
-    sct_builtin  = 1 << 1,  /* search builtins */
-    sct_function = 1 << 2,  /* search functions */
-    sct_defpath  = 1 << 3,  /* search the default PATH */
-    sct_rbpath   = 1 << 4,  /* return path of regular builtin */
-    sct_noslash  = 1 << 5,  /* don't check if name contains slash */
+    SCT_EXTERNAL = 1 << 0,  /* search for an external command */
+    SCT_BUILTIN  = 1 << 1,  /* search for a built-in */
+    SCT_FUNCTION = 1 << 2,  /* search for a function */
+    SCT_ALL      = 1 << 3,  /* search all */
+    SCT_STDPATH  = 1 << 4,  /* search the standard PATH */
+    SCT_NOSLASH  = 1 << 5,  /* assume command name contains no slash */
 };
 
 /* info about a simple command to execute */
@@ -677,8 +677,13 @@ void exec_commands(command_T *c, exec_T type)
     }
 
     handle_signals();
-    if (shopt_errexit && !supresserrexit && laststatus != Exit_SUCCESS
-	    && lasttype == CT_SIMPLE)
+
+    if (shopt_errexit && !supresserrexit
+	    && laststatus != Exit_SUCCESS && lasttype == CT_SIMPLE
+#if YASH_ENABLE_LINEEDIT
+	    && !(le_state & LE_STATE_COMPLETING)
+#endif
+	    )
 	exit_shell_with_status(laststatus);
 }
 
@@ -772,9 +777,9 @@ pid_t exec_process(
     commandinfo_T cmdinfo;
     bool temp;
 
-    /* First, we check if the command is a special builtin or a function
+    /* First, we check if the command is a special built-in or a function
      * and determine whether we have to open a temporary environment. */
-    search_command(argv0, argv[0], &cmdinfo, sct_builtin | sct_function);
+    search_command(argv0, argv[0], &cmdinfo, SCT_BUILTIN | SCT_FUNCTION);
     special_builtin_executed = (cmdinfo.type == specialbuiltin);
     temp = c->c_assigns && assignment_is_temporary(cmdinfo.type);
     if (temp)
@@ -796,7 +801,7 @@ pid_t exec_process(
     if (cmdinfo.type == externalprogram) {
 	if (posixly_correct) {
 	    search_command(argv0, argv[0], &cmdinfo,
-		    sct_external | sct_builtin);
+		    SCT_EXTERNAL | SCT_BUILTIN);
 	} else {
 	    if (wcschr(argv[0], L'/')) {
 		cmdinfo.ci_path = argv0;
@@ -805,7 +810,7 @@ pid_t exec_process(
 			goto done3;
 	    } else {
 		search_command(argv0, argv[0], &cmdinfo,
-			sct_external | sct_builtin | sct_noslash);
+			SCT_EXTERNAL | SCT_BUILTIN | SCT_NOSLASH);
 		if (cmdinfo.type == externalprogram && cmdinfo.ci_path == NULL)
 		    if (command_not_found_handler(argv))
 			goto done3;
@@ -922,72 +927,78 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
     return cpid;
 }
 
-/* Searches and determines the command to execute.
+/* Searches for a command.
  * The result is assigned to `*ci'.
  * If the command is not found, `ci->type' is set to `externalprogram' and
  * `ci->ci_path' is set to NULL.
- * `name' and `wname' must contain the same string value. */
-/* When the result type `ci->type' is `regularbuiltin', the result value is
- * either a function pointer to the builtin (if the `sct_rbpath' flag is not
- * specified in `type') or the path of the corresponding external command
- * (otherwise). */
-/* In the POSIXly correct mode, you need the sct_external flag to find a regular
- * builtin. */
+ * `name' and `wname' must contain the same string value.
+ * If the SCT_ALL flag is not set:
+ *   *  a function whose name contains a slash cannot be found
+ *   *  a regular built-in cannot be found in the POSIXly correct mode if the
+ *      SCT_EXTERNAL flag is not set either. */
 void search_command(
 	const char *restrict name, const wchar_t *restrict wname,
 	commandinfo_T *restrict ci, enum srchcmdtype_T type)
 {
+    bool slash = !(type & SCT_NOSLASH) && wcschr(wname, L'/');
+
     const builtin_T *bi;
-    command_T *funcbody;
-
-    if (!(type & sct_noslash) && wcschr(wname, L'/')) {
-	if (!(type & sct_external))
-	    goto notfound;
-	ci->type = externalprogram;
-	ci->ci_path = name;
-	return;
-    }
-
-    /* search builtins and functions. */
-    bi = (type & sct_builtin) ? get_builtin(name) : NULL;
+    if (!slash && (type & SCT_BUILTIN))
+	bi = get_builtin(name);
+    else
+	bi = NULL;
     if (bi && bi->type == BI_SPECIAL) {
 	ci->type = specialbuiltin;
 	ci->ci_builtin = bi->body;
 	return;
     }
-    funcbody = (type & sct_function) ? get_function(wname) : NULL;
-    if (funcbody) {
-	ci->type = function;
-	ci->ci_function = funcbody;
-	return;
-    } else if (bi && bi->type == BI_SEMISPECIAL) {
-	ci->type = semispecialbuiltin;
-	ci->ci_builtin = bi->body;
-	return;
-    } else if (bi && !posixly_correct && !(type & sct_rbpath)) {
-	ci->type = regularbuiltin;
-	ci->ci_builtin = bi->body;
-	return;
+
+    if ((type & SCT_FUNCTION) && (!slash || (type & SCT_ALL))) {
+	command_T *funcbody = get_function(wname);
+	if (funcbody) {
+	    ci->type = function;
+	    ci->ci_function = funcbody;
+	    return;
+	}
     }
 
-    if (!(type & (sct_external | sct_rbpath)))
-	goto notfound;
-
-    ci->ci_path = (type & sct_defpath)
-	? get_command_path_default(name) : get_command_path(name, false);
-    if (bi && (ci->ci_path || !posixly_correct)) {
-	assert(bi->type == BI_REGULAR);
-	ci->type = regularbuiltin;
-	if (!(type & sct_rbpath))
+    if (bi) {
+	if (bi->type == BI_SEMISPECIAL) {
+	    ci->type = semispecialbuiltin;
 	    ci->ci_builtin = bi->body;
-    } else {
-	ci->type = externalprogram;
-	if (!(type & sct_external))
-	    ci->ci_path = NULL;
+	    return;
+	} else if (!posixly_correct) {
+	    goto regular_builtin;
+	}
     }
-    return;
 
-notfound:
+    if (slash && (type & SCT_EXTERNAL)) {
+	ci->type = externalprogram;
+	ci->ci_path = name;
+	return;
+    }
+
+    if ((type & SCT_EXTERNAL) || (bi && (type & SCT_ALL))) {
+	const char *cmdpath;
+	if (type & SCT_STDPATH)
+	    cmdpath = get_command_path_default(name);
+	else
+	    cmdpath = get_command_path(name, false);
+	if (cmdpath) {
+	    if (bi) {
+regular_builtin:
+		assert(bi->type == BI_REGULAR);
+		ci->type = regularbuiltin;
+		ci->ci_builtin = bi->body;
+	    } else {
+		ci->type = externalprogram;
+		ci->ci_path = cmdpath;
+	    }
+	    return;
+	}
+    }
+
+    /* command not found... */
     ci->type = externalprogram;
     ci->ci_path = NULL;
     return;
@@ -1169,7 +1180,7 @@ void print_xtrace(void *const *argv)
 
     if (shopt_xtrace && (tracevars || argv != NULL)
 #if YASH_ENABLE_LINEEDIT
-	    && !le_state_is_active
+	    && !(le_state & LE_STATE_ACTIVE)
 #endif
 	    ) {
 	bool first = true;
@@ -1428,47 +1439,94 @@ int exec_variable_as_commands(const wchar_t *varname, const char *codename)
 
 #if YASH_ENABLE_LINEEDIT
 
-/* Generates completion candidates by executing the specified function.
- * Returns the exit status of the function. */
-/* The prototype of this function is declared in "lineedit/complete.h". */
-int generate_candidates_using_function(
-	const wchar_t *funcname, le_context_T *context)
+/* Autoloads the specified file to load a completion function definition.
+ * String `cmdname', which may be NULL, is used as the only positional parameter
+ * during script execution.
+ * Returns true if a file was autoloaded. */
+bool autoload_completion_function_file(
+	const wchar_t *filename, const wchar_t *cmdname)
 {
-    if (funcname == NULL)
-	return Exit_SUCCESS;
+    char *mbsfilename = malloc_wcstombs(filename);
+    if (mbsfilename == NULL)
+	return false;
 
-    le_compdebug("executing candidate generator function \"%ls\"", funcname);
+    char *path = which(mbsfilename, get_path_array(PA_LOADPATH),
+	    is_readable_regular);
+    if (path == NULL) {
+	le_compdebug("file \"%s\" was not found in $YASH_LOADPATH",
+		mbsfilename);
+	free(mbsfilename);
+	return false;
+    }
 
-    command_T *function = get_function(funcname);
-    if (function == NULL) {
-	le_compdebug("  -- no such function");
-	return Exit_NOTFOUND;
+    int fd = move_to_shellfd(open(path, O_RDONLY));
+    if (fd < 0) {
+	le_compdebug("cannot open the file \"%s\"", path);
+	free(path);
+	return false;
     }
 
     struct parsestate_T *state = save_parse_state();
+    int savelaststatus = laststatus;
+    bool saveposix = posixly_correct;
+    posixly_correct = false;
+    open_new_environment(false);
+    set_positional_parameters((void *[]) { (void *) cmdname, NULL });
 
-    void *args[context->pwordc + 2];
-    memcpy(args, context->pwords, context->pwordc * sizeof *args);
-    args[context->pwordc] = context->origsrc;
-    args[context->pwordc + 1] = NULL;
+    le_compdebug("executing file \"%s\" (autoload)", path);
+    exec_input(fd, mbsfilename, false, false, false);
+    le_compdebug("finished executing file \"%s\"", path);
 
-    int savelaststatus = laststatus, resultstatus;
+    close_current_environment();
+    posixly_correct = saveposix;
+    laststatus = savelaststatus;
+    restore_parse_state(state);
+    remove_shellfd(fd);
+    xclose(fd);
+    free(path);
+    free(mbsfilename);
+    return true;
+}
+
+/* Calls the function whose name is `funcname' as a completion function.
+ * Returns false if no such function has been defined. */
+bool call_completion_function(const wchar_t *funcname)
+{
+    command_T *func = get_function(funcname);
+    if (func == NULL) {
+	le_compdebug("completion function \"%ls\" is not defined", funcname);
+	return false;
+    }
+
+    struct parsestate_T *state = save_parse_state();
+    int savelaststatus = laststatus;
     bool saveposix = posixly_correct;
     posixly_correct = false;
 
-    exec_function_body(function, args, false);
+    le_compdebug("executing completion function \"%ls\"", funcname);
+
+    /* don't use `exec_function_body': We have to prepare special local
+     * variables $WORDS and $TARGETWORD. */
+    savefd_T *savefd;
+    open_new_environment(false);
+    set_positional_parameters((void *[]) { NULL });
+    set_completion_variables();
+    if (open_redirections(func->c_redirs, &savefd)) {
+	exec_nonsimple_command(func, false);
+	if (execinfo.exception == ee_return)
+	    execinfo.exception = ee_none;
+    }
+    undo_redirections(savefd);
+    close_current_environment();
+
+    le_compdebug("finished executing completion function \"%ls\"", funcname);
+    le_compdebug("  with the exit status of %d", laststatus);
 
     posixly_correct = saveposix;
-    resultstatus = laststatus;
     laststatus = savelaststatus;
-
-    le_compdebug("finished executing function \"%ls\"", funcname);
-    if (resultstatus != Exit_SUCCESS)
-	le_compdebug("function returned exit status of %d", resultstatus);
-
     restore_parse_state(state);
 
-    return resultstatus;
+    return true;
 }
 
 #endif /* YASH_ENABLE_LINEEDIT */
@@ -1482,17 +1540,17 @@ static int command_builtin_execute(
 	int argc, void **argv, enum srchcmdtype_T type)
     __attribute__((nonnull));
 static bool print_command_info(const wchar_t *commandname,
-	enum srchcmdtype_T type, bool keywordalias, bool humanfriendly)
+	enum srchcmdtype_T type, bool alias, bool keyword, bool humanfriendly)
     __attribute__((nonnull));
-static void print_command_absolute_path(
+static bool print_command_absolute_path(
 	const char *name, const char *path, bool humanfriendly)
     __attribute__((nonnull));
 
 /* Options for the "break", "continue" and "eval" builtins. */
 static const struct xoption iter_options[] = {
-    { L"iteration", xno_argument, L'i', },
+    { L"iteration", OPTARG_NONE, L'i', },
 #if YASH_ENABLE_HELP
-    { L"help",      xno_argument, L'-', },
+    { L"help",      OPTARG_NONE, L'-', },
 #endif
     { NULL, 0, 0, },
 };
@@ -1701,28 +1759,33 @@ const char eval_help[] = Ngt(
 #endif
 
 /* The "." builtin, which accepts the following option:
- *  -A: disable aliases */
+ *  -A: disable aliases
+ *  -L: autoload */
 int dot_builtin(int argc, void **argv)
 {
     static const struct xoption long_options[] = {
-	{ L"no-alias", xno_argument, L'A', },
+	{ L"no-alias", OPTARG_NONE, L'A', },
+	{ L"autoload", OPTARG_NONE, L'L', },
 #if YASH_ENABLE_HELP
-	{ L"help",     xno_argument, L'-', },
+	{ L"help",     OPTARG_NONE, L'-', },
 #endif
 	{ NULL, 0, 0, },
     };
 
-    bool enable_alias = true;
+    bool enable_alias = true, autoload = false;
 
     wchar_t opt;
     xoptind = 0, xopterr = true;
     while ((opt = xgetopt_long(argv,
-		    posixly_correct ? L"+" : L"+A",
+		    posixly_correct ? L"+" : L"+AL",
 		    long_options, NULL))) {
 	switch (opt) {
 	    case L'A':
 		enable_alias = false;
 		break;
+	    case L'L':
+		autoload = true;
+		break;
 #if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
@@ -1730,7 +1793,7 @@ int dot_builtin(int argc, void **argv)
 	    default:  print_usage:
 		fprintf(stderr, gt(posixly_correct
 			    ? Ngt("Usage:  . file\n")
-			    : Ngt("Usage:  . [-A] file [arg...]\n")));
+			    : Ngt("Usage:  . [-AL] file [arg...]\n")));
 		SPECIAL_BI_ERROR;
 		return Exit_ERROR;
 	}
@@ -1751,7 +1814,17 @@ int dot_builtin(int argc, void **argv)
     }
 
     char *path;
-    if (!wcschr(filename, L'/')) {
+    if (autoload) {
+	path = which(mbsfilename, get_path_array(PA_LOADPATH),
+		is_readable_regular);
+	if (path == NULL) {
+	    xerror(0, Ngt("%s: not found in $YASH_LOADPATH"), mbsfilename);
+	    free(mbsfilename);
+	    if (!is_interactive)
+		exit_shell_with_status(Exit_FAILURE);
+	    return Exit_FAILURE;
+	}
+    } else if (!wcschr(filename, L'/')) {
 	path = which(mbsfilename, get_path_array(PA_PATH), is_readable_regular);
 	if (!path) {
 	    if (!posixly_correct) {
@@ -1799,16 +1872,19 @@ int dot_builtin(int argc, void **argv)
 #if YASH_ENABLE_HELP
 const char dot_help[] = Ngt(
 "dot - read file and execute commands\n"
-"\t. [-A] file [arg...]\n"
+"\t. [-AL] file [arg...]\n"
 "Reads the specified <file> and executes commands in it.\n"
 "If <arg>s are specified, they are used as the positional parameters.\n"
 "Otherwise, the positional parameters are not changed.\n"
 "If <file> does not contain any slashes, the shell searches $PATH for a\n"
 "readable shell script file whose name is <file>. To ensure that the file in\n"
 "the current working directory is used, start <file> with \"./\".\n"
+"If the -L (--autoload) option is specified, the shell searches\n"
+"$YASH_LOADPATH instead of $PATH, regardless of whether <file> contains\n"
+"slashes.\n"
 "If the -A (--no-alias) option is specified, alias substitution is not\n"
 "performed during processing the file.\n"
-"In POSIXly correct mode, the -A option cannot be used and <arg>s must not be\n"
+"In POSIXly correct mode, options cannot be used and <arg>s must not be\n"
 "given.\n"
 );
 #endif
@@ -1820,11 +1896,11 @@ const char dot_help[] = Ngt(
 int exec_builtin(int argc, void **argv)
 {
     static const struct xoption long_options[] = {
-	{ L"as",    xrequired_argument, L'a', },
-	{ L"clear", xno_argument,       L'c', },
-	{ L"force", xno_argument,       L'f', },
+	{ L"as",    OPTARG_REQUIRED, L'a', },
+	{ L"clear", OPTARG_NONE,     L'c', },
+	{ L"force", OPTARG_NONE,     L'f', },
 #if YASH_ENABLE_HELP
-	{ L"help",  xno_argument,       L'-', },
+	{ L"help",  OPTARG_NONE,     L'-', },
 #endif
 	{ NULL, 0, 0, },
     };
@@ -1984,38 +2060,48 @@ const char exec_help[] = Ngt(
 #endif
 
 /* The "command"/"type" builtin, which accepts the following options:
+ *  -a: search aliases
  *  -b: search builtins
- *  -B: search external commands
+ *  -e: search external commands
+ *  -f: search functions
+ *  -k: search keywords
  *  -p: use the default path to find the command
  *  -v: print info about the command
  *  -V: print info about the command in a human-friendly format */
 int command_builtin(int argc, void **argv)
 {
     static const struct xoption long_options[] = {
-	{ L"builtin-only",     xno_argument, L'b', },
-	{ L"external-only",    xno_argument, L'B', },
-	{ L"standard-path",    xno_argument, L'p', },
-	{ L"identify",         xno_argument, L'v', },
-	{ L"verbose-identify", xno_argument, L'V', },
+	{ L"alias",            OPTARG_NONE, L'a', },
+	{ L"builtin-command",  OPTARG_NONE, L'b', },
+	{ L"external-command", OPTARG_NONE, L'e', },
+	{ L"function",         OPTARG_NONE, L'f', },
+	{ L"keyword",          OPTARG_NONE, L'k', },
+	{ L"standard-path",    OPTARG_NONE, L'p', },
+	{ L"identify",         OPTARG_NONE, L'v', },
+	{ L"verbose-identify", OPTARG_NONE, L'V', },
 #if YASH_ENABLE_HELP
-	{ L"help",             xno_argument, L'-', },
+	{ L"help",             OPTARG_NONE, L'-', },
 #endif
 	{ NULL, 0, 0, },
     };
 
     bool argv0istype = wcscmp(ARGV(0), L"type") == 0;
     bool printinfo = argv0istype, humanfriendly = argv0istype;
-    bool builtin = false, external = false, defpath = false;
+    enum srchcmdtype_T type = 0;
+    bool aliases = false, keywords = false, defpath = false;
 
     wchar_t opt;
     xoptind = 0, xopterr = true;
     while ((opt = xgetopt_long(argv,
-		    posixly_correct ? L"+pvV" : L"+bBpvV",
+		    posixly_correct ? L"+pvV" : L"+abefkpvV",
 		    long_options, NULL))) {
 	switch (opt) {
-	    case L'b':  builtin   = true;  break;
-	    case L'B':  external  = true;  break;
-	    case L'p':  defpath   = true;  break;
+	    case L'a':  aliases = true;        break;
+	    case L'b':  type |= SCT_BUILTIN;   break;
+	    case L'e':  type |= SCT_EXTERNAL;  break;
+	    case L'f':  type |= SCT_FUNCTION;  break;
+	    case L'k':  keywords = true;       break;
+	    case L'p':  defpath = true;        break;
 	    case L'v':  printinfo = true;  humanfriendly = false;  break;
 	    case L'V':  printinfo = true;  humanfriendly = true;   break;
 #if YASH_ENABLE_HELP
@@ -2027,42 +2113,42 @@ int command_builtin(int argc, void **argv)
 	}
     }
 
-    if (xoptind == argc) {
-	if (posixly_correct)
-	    goto print_usage;
-	else
-	    return Exit_SUCCESS;
-    } else if (posixly_correct && (builtin || external)) {
-	goto print_usage;
-    }
-
-    enum srchcmdtype_T type = 0;
-    if (builtin)
-	type |= sct_builtin;
-    if (external)
-	type |= sct_external;
-    if (!builtin && !external)
-	type |= sct_builtin | sct_external;
-    if (defpath)
-	type |= sct_defpath;
-
     if (!printinfo) {
+	if (aliases || keywords)
+	    goto print_usage;
+	if (xoptind == argc) {
+	    if (posixly_correct)
+		goto print_usage;
+	    else
+		return Exit_SUCCESS;
+	}
+	if (!type)
+	    type = SCT_EXTERNAL | SCT_BUILTIN;
+	else
+	    type |= SCT_ALL;
+	if (defpath)
+	    type |= SCT_STDPATH;
 	return command_builtin_execute(
 		argc - xoptind, argv + xoptind, type);
     } else {
-	bool keywordalias = false;
-	if (!builtin && !external)
-	    type |= sct_function, keywordalias = true;
+	if (!argv0istype && posixly_correct && argc - xoptind != 1)
+	    goto print_usage;
+	if (!type && !aliases && !keywords) {
+	    type = SCT_EXTERNAL | SCT_BUILTIN | SCT_FUNCTION;
+	    aliases = keywords = true;
+	} else {
+	    type |= SCT_ALL;
+	}
+	if (defpath)
+	    type |= SCT_STDPATH;
 
 	bool ok = true;
 	clearerr(stdout);
 	for (int i = xoptind; i < argc; i++) {
 	    ok &= print_command_info(
-		    ARGV(i), type, keywordalias, humanfriendly);
-	    if (ferror(stdout)) {
-		xerror(0, Ngt("cannot print to standard output"));
-		return Exit_FAILURE;
-	    }
+		    ARGV(i), type, aliases, keywords, humanfriendly);
+	    if (ferror(stdout))
+		break;
 	}
 	return ok ? Exit_SUCCESS : Exit_FAILURE;
     }
@@ -2072,10 +2158,10 @@ print_usage:
 	fprintf(stderr, gt("Usage:  type command...\n"));
     else if (posixly_correct)
 	fprintf(stderr, gt("Usage:  command [-p] command [arg...]\n"
-			   "        command -v|-V [-p] command...\n"));
+			   "        command -v|-V [-p] command\n"));
     else
-	fprintf(stderr, gt("Usage:  command [-bBp] command [arg...]\n"
-			   "        command -v|-V [-bBp] command...\n"));
+	fprintf(stderr, gt("Usage:  command [-befp] command [arg...]\n"
+			   "        command -v|-V [-abefkp] command...\n"));
     return Exit_ERROR;
 }
 
@@ -2088,8 +2174,11 @@ int command_builtin_execute(int argc, void **argv, enum srchcmdtype_T type)
     commandinfo_T ci;
     bool finally_exit = false;
 
-    if (!argv0)
-	argv0 = xstrdup("");
+    if (argv0 == NULL) {
+	xerror(EILSEQ, Ngt("invalid command name"));
+	return Exit_NOTFOUND;
+    }
+
     search_command(argv0, argv[0], &ci, type);
     if (ci.type == externalprogram) {
 	pid_t cpid = fork_and_reset(0, true, t_leave);
@@ -2114,106 +2203,118 @@ int command_builtin_execute(int argc, void **argv, enum srchcmdtype_T type)
 }
 
 /* Prints info about the specified command.
- * If the command is not found, returns false.
- * The shell keywords and aliases are printed only when `keywordalias' is true.
- */
+ * If the command is not found, returns false. */
 bool print_command_info(
 	const wchar_t *commandname, enum srchcmdtype_T type,
-	bool keywordalias, bool humanfriendly)
+	bool aliases, bool keywords, bool humanfriendly)
 {
-    bool found;
+    const char *msgfmt;
+    char *name = NULL;
+    bool found = false;
 
-    if (keywordalias && is_keyword(commandname)) {
-	printf(humanfriendly ? gt("%ls: shell keyword\n") : "%ls\n",
-		commandname);
+    if (keywords && is_keyword(commandname)) {
+	msgfmt = humanfriendly ? gt("%ls: shell keyword\n") : "%ls\n";
+	if (printf(msgfmt, commandname) < 0)
+	    goto ioerror;
 	return true;
     }
+
 #if YASH_ENABLE_ALIAS
-    found = keywordalias && print_alias_if_defined(commandname, humanfriendly);
-    if (found && !humanfriendly)
-	return true;
+    if (aliases) {
+	if (print_alias_if_defined(commandname, humanfriendly)) {
+	    return true;
+	} else {
+	    if (ferror(stdout))
+		return false;
+	}
+    }
 #else
-    found = false;
+    (void) aliases;
 #endif
 
-    char *name = malloc_wcstombs(commandname);
-    if (!name)
+    name = malloc_wcstombs(commandname);
+    if (name == NULL)
 	return false;
 
     commandinfo_T ci;
-    search_command(name, commandname, &ci, type | sct_rbpath);
+    search_command(name, commandname, &ci, type);
     switch (ci.type) {
 	case externalprogram:
 	    if (ci.ci_path && is_executable_regular(ci.ci_path)) {
-		print_command_absolute_path(name, ci.ci_path, humanfriendly);
-		goto ok;
-	    } else if (found) {
-		goto ok;
-	    } else {
-		if (humanfriendly)
-		    xerror(0, gt("%s: no such command or function"), name);
-		goto error;
+		if (!print_command_absolute_path(
+			    name, ci.ci_path, humanfriendly))
+		    goto ioerror;
+		found = true;
 	    }
+	    break;
 	case specialbuiltin:
-	    if (humanfriendly)
-		printf(gt("%s: special builtin\n"), name);
-	    else
-		puts(name);
-	    goto ok;
+	    msgfmt = humanfriendly ? gt("%s: special builtin\n") : "%s\n";
+	    if (printf(msgfmt, name) < 0)
+		goto ioerror;
+	    found = true;
+	    break;
 	case semispecialbuiltin:
-	    if (humanfriendly)
-		printf(gt("%s: semi-special builtin\n"), name);
+	    msgfmt = humanfriendly ? gt("%s: semi-special builtin\n") : "%s\n";
+	    if (printf(msgfmt, name) < 0)
+		goto ioerror;
+	    found = true;
+	    break;
+	case regularbuiltin:;
+	    const char *cmdpath;
+	    if (type & SCT_STDPATH)
+		cmdpath = get_command_path_default(name);
 	    else
-		puts(name);
-	    goto ok;
-	case regularbuiltin:
-	    if (ci.ci_path) {
-		if (humanfriendly)
-		    printf(gt("%s: regular builtin at %s\n"), name, ci.ci_path);
+		cmdpath = get_command_path(name, false);
+	    if (humanfriendly) {
+		if (cmdpath == NULL)
+		    msgfmt = gt("%s: regular builtin (not found in $PATH)\n");
 		else
-		    puts(ci.ci_path);
+		    msgfmt = gt("%s: regular builtin at %s\n");
+		if (printf(msgfmt, name, cmdpath) < 0)
+		    goto ioerror;
 	    } else {
-		if (humanfriendly)
-		    printf(gt("%s: regular builtin (not found in $PATH)\n"),
-			    name);
-		else
-		    puts(name);
+		if (puts(cmdpath == NULL ? name : cmdpath) < 0)
+		    goto ioerror;
 	    }
-	    goto ok;
+	    found = true;
+	    break;
 	case function:
-	    if (humanfriendly)
-		printf(gt("%s: function\n"), name);
-	    else
-		puts(name);
-	    goto ok;
+	    msgfmt = humanfriendly ? gt("%s: function\n") : "%s\n";
+	    if (printf(msgfmt, name) < 0)
+		goto ioerror;
+	    found = true;
+	    break;
     }
 
-ok:
+    if (!found && humanfriendly)
+	xerror(0, Ngt("%s: no such command or function"), name);
     free(name);
-    return true;
-error:
+    return found;
+
+ioerror:
+    xerror(errno, Ngt("cannot print to standard output"));
     free(name);
     return false;
 }
 
-/* Prints the absolute path of the specified command. */
-void print_command_absolute_path(
+/* Prints the absolute path of the specified command.
+ * Returns false on error (errno is set). */
+bool print_command_absolute_path(
 	const char *name, const char *path, bool humanfriendly)
 {
-    if (path[0] == L'/') {
+    if (path[0] == '/') {
 	/* the path is already absolute */
 	if (humanfriendly)
-	    printf(gt("%s: external command at %s\n"), name, path);
+	    return printf(gt("%s: external command at %s\n"), name, path) >= 0;
 	else
-	    puts(path);
-	return;
+	    return puts(path) >= 0;
     }
 
     const wchar_t *wpwd = getvar(L VAR_PWD);
     char *pwd = NULL;
     if (wpwd) {
 	pwd = malloc_wcstombs(wpwd);
-	if (!is_same_file(pwd, ".")) {
+	if (pwd && !is_same_file(pwd, ".")) {
 	    free(pwd);
 	    pwd = NULL;
 	}
@@ -2224,37 +2325,44 @@ void print_command_absolute_path(
 	    pwd = xstrdup(".");  /* last resort */
     }
 
+    int r, saveerrno;
     if (humanfriendly)
-	printf(gt("%s: external command at %s/%s\n"), name, pwd, path);
+	r = printf(gt("%s: external command at %s/%s\n"), name, pwd, path);
     else
-	printf("%s/%s\n", pwd, path);
+	r = printf("%s/%s\n", pwd, path);
+    saveerrno = errno;
     free(pwd);
-    return;
+    errno = saveerrno;
+    return r >= 0;
 }
 
 #if YASH_ENABLE_HELP
 
 const char command_help[] = Ngt(
 "command - execute or identify command\n"
-"\tcommand [-bBp] command [argument...]\n"
-"\tcommand -v|-V [-bBp] command...\n"
+"\tcommand [-befp] command [argument...]\n"
+"\tcommand -v|-V [-abefkp] command...\n"
 "Executes or identifies the specified command.\n"
-"Without the -v or -V option, <command> is executed with given <argument>s if\n"
-"any. <command> is treated as a builtin or external command, but not a\n"
-"function or an alias. If the -p (--standard-path) option is given, the\n"
-"system's default PATH is searched for the command instead of the current\n"
-"$PATH. If the -b (--builtin-only) option (but not -B) is given, <command> is\n"
-"treated as a builtin only: the external command is ignored. The -B\n"
-"(--external-only) option does the opposite: an external command is executed\n"
-"ignoring the builtin.\n"
-"With the -v (--identify) option, identifies <command>. If the command is\n"
-"found in $PATH, its full path is printed. If it is a builtin or a function,\n"
+"Without the -v or -V option, <command> is executed with <argument>s given.\n"
+"<command> is treated as a built-in or external command or a function\n"
+"according to the options specified. If the -p (--standard-path) option is\n"
+"given, the system's default PATH is searched for the command instead of the\n"
+"current $PATH.\n"
+"With the -v (--identify) option, <command> is identified. If the command is\n"
+"found in $PATH, its full path is printed. If it is a built-in or a function,\n"
 "the command name is simply printed. If it is an alias, it is printed in the\n"
 "form like \"alias ll='ls -l'\". If the command is not found, nothing is\n"
 "printed and the exit status is non-zero.\n"
 "With the -V (--verbose-identify) option, the command is identified in the\n"
 "same way but the result is printed verbosely in a human-readable form.\n"
-"In the POSIXly correct mode, the -b and -B options are not available.\n"
+"The following options specify the command type:\n"
+"  -a --alias\n"
+"  -b --builtin-command\n"
+"  -e --external-command\n"
+"  -f --function\n"
+"  -k --keyword\n"
+"When none of these are specified, they default to -be (without -v or -V) or\n"
+"-abefk (with -v or -V)\n"
 );
 
 const char type_help[] = Ngt(
