@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* input.c: functions for input of command line */
-/* (C) 2007-2009 magicant */
+/* (C) 2007-2010 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -86,48 +86,80 @@ inputresult_T input_wcs(struct xwcsbuf_T *buf, void *inputinfo)
 }
 
 /* An input function that reads input from a file stream.
- * `inputinfo' is a pointer to a `FILE', which must be set to non-blocking.
- * Reads one line with `fgetws' and appends it to the buffer. */
+ * `inputinfo' is a pointer to a `struct input_file_info'.
+ * Reads one line from `inputinfo->fd' and appends it to the buffer. */
 inputresult_T input_file(struct xwcsbuf_T *buf, void *inputinfo)
 {
-    FILE *f = inputinfo;
-    int fd = fileno(f);
+    return read_input(buf, inputinfo, true);
+}
+
+/* Reads one line from file descriptor `info->fd' and appends it to `buf'.
+ * If `trap' is true, traps are handled while reading and the `sigint_received'
+ * flag is cleared when this function returns.
+ * Returns:
+ *   INPUT_OK    if at least one character was appended
+ *   INPUT_EOF   if reached end of file without reading any characters
+ *   INPUT_ERROR if an error occurred before reading any characters */
+inputresult_T read_input(
+	xwcsbuf_T *buf, struct input_file_info *info, bool trap)
+{
     size_t initlen = buf->length;
     bool ok = true;
 
-    handle_signals();
-    while (ok) {
-	wb_ensuremax(buf, buf->length + 100);
-	if (fgetws(buf->contents + buf->length,
-		    buf->maxlength - buf->length, f)) {
-	    size_t len = wcslen(buf->contents + buf->length);
-	    // `len' may be 0 if a null character is input
-	    buf->length += len;
-	    if (len == 0 || buf->contents[buf->length - 1] == L'\n')
+    if (trap) {
+	handle_signals();
+	reset_sigint();
+    }
+
+    for (;;) {
+	if (info->bufpos >= info->bufmax) {
+read_input:  /* if there's nothing in the buffer, read the next input */
+	    if (!(ok = wait_for_input(info->fd, trap, -1)))
 		goto end;
-	} else {
-	    buf->contents[buf->length] = L'\0';
-	    if (feof(f)) {
-		clearerr(f);
-		goto end;
-	    }
-	    assert(ferror(f));
-	    switch (errno) {
+
+	    ssize_t readcount = read(info->fd, info->buf, info->bufsize);
+	    if (readcount < 0) switch (errno) {
 		case EINTR:
 		case EAGAIN:
 #if EAGAIN != EWOULDBLOCK
 		case EWOULDBLOCK:
 #endif
-		    clearerr(f);
-		    ok = wait_for_input(fd, true, -1);
-		    break;
+		    goto read_input;  /* try again */
 		default:
-		    xerror(errno, Ngt("cannot read input"));
-		    ok = false;
-		    break;
+		    goto error;
+	    } else if (readcount == 0) {
+		goto end;
 	    }
+	    info->bufpos = 0;
+	    info->bufmax = readcount;
+	}
+
+	/* convert bytes in `info->buf' into a wide character and
+	 * append it to `buf' */
+	wb_ensuremax(buf, buf->length + 1);
+	assert(info->bufpos < info->bufmax);
+	size_t convcount = mbrtowc(buf->contents + buf->length,
+		info->buf + info->bufpos, info->bufmax - info->bufpos,
+		&info->state);
+	switch (convcount) {
+	    case 0:            /* read null character */
+		goto end;
+	    case (size_t) -1:  /* not a valid character */
+		goto error;
+	    case (size_t) -2:  /* needs more input */
+		goto read_input;
+	    default:
+		info->bufpos += convcount;
+		buf->contents[++buf->length] = L'\0';
+		if (buf->contents[buf->length - 1] == L'\n')
+		    goto end;
+		break;
 	}
     }
+
+error:
+    xerror(errno, Ngt("cannot read input"));
+    ok = false;
 end:
     if (initlen != buf->length)
 	return INPUT_OK;
@@ -137,92 +169,12 @@ end:
 	return INPUT_ERROR;
 }
 
-/* An input function that reads input from the standard input.
- * Bytes are read one by one until a newline is encountered. No more bytes are
- * read after the newline.
- * This function does not use `inputinfo'.
- * The result is appended to the buffer. */
-inputresult_T input_stdin(
-	struct xwcsbuf_T *buf, void *inputinfo __attribute__((unused)))
-{
-    size_t initlen = buf->length;
-    bool ok = read_line_from_stdin(buf, true);
-    if (initlen != buf->length)
-	return INPUT_OK;
-    else if (ok)
-	return INPUT_EOF;
-    else
-	return INPUT_ERROR;
-}
-
-/* Reads a line of input from the standard input.
- * Bytes are read one by one until a newline is encountered. No more bytes are
- * read after the newline.
- * If `trap' is true, traps are handled while reading.
- * The result is appended to the buffer.
- * Returns true iff successful. */
-bool read_line_from_stdin(struct xwcsbuf_T *buf, bool trap)
-{
-    static bool initialized = false;
-    static mbstate_t state;
-
-    bool ok = true;
-
-    if (!initialized) {
-	initialized = true;
-	memset(&state, 0, sizeof state);  /* initialize the state */
-    }
-
-    handle_signals();
-    set_nonblocking(STDIN_FILENO);
-    while (ok) {
-	char c;
-	ssize_t n = read(STDIN_FILENO, &c, 1);
-	if (n < 0) switch (errno) {
-	    case EINTR:
-	    case EAGAIN:
-#if EAGAIN != EWOULDBLOCK
-	    case EWOULDBLOCK:
-#endif
-		ok = wait_for_input(STDIN_FILENO, trap, -1);
-		break;
-	    default:
-		xerror(errno, Ngt("cannot read input"));
-		ok = false;
-		break;
-	} else if (n == 0) {
-	    goto done;
-	} else {
-	    wchar_t wc;
-	    switch (mbrtowc(&wc, &c, 1, &state)) {
-		case 0:
-		case 1:
-		    wb_wccat(buf, wc);
-		    if (wc == L'\n')
-			goto done;
-		    break;
-		case (size_t) -2:
-		    break;
-		case (size_t) -1:
-		    xerror(errno, Ngt("cannot read input"));
-		    ok = false;
-		    break;
-		default:
-		    assert(false);
-	    }
-	}
-    }
-done:
-    unset_nonblocking(STDIN_FILENO);
-    return ok;
-}
-
 /* An input function that prints a prompt and reads input.
- * `inputinfo' is a pointer to a `struct input_readline_info'.
+ * `inputinfo' is a pointer to a `struct input_interactive_info'.
  * `inputinfo->type' must be 1 or 2, specifying the type of the prompt.
  * For example, $PS1 is printed if `inputinfo->type' is 1.
  * If `inputinfo->type' is 1, this function changes it to 2. */
-inputresult_T input_readline(struct xwcsbuf_T *buf, void *inputinfo)
+inputresult_T input_interactive(struct xwcsbuf_T *buf, void *inputinfo)
 {
 #if YASH_ENABLE_LINEEDIT
     /* An input function must not return more than one line at a time.
@@ -236,17 +188,16 @@ inputresult_T input_readline(struct xwcsbuf_T *buf, void *inputinfo)
 #endif
 
     struct parsestate_T *state = save_parse_state();
-    struct input_readline_info *info = inputinfo;
+    struct input_interactive_info *info = inputinfo;
     struct promptset_T prompt;
 
-    if (info->type == 1)
+    if (info->prompttype == 1)
 	if (!posixly_correct)
-	    exec_variable_as_commands(
-		    L VAR_PROMPT_COMMAND, VAR_PROMPT_COMMAND);
-    prompt = get_prompt(info->type);
+	    exec_variable_as_commands(L VAR_PROMPT_COMMAND, VAR_PROMPT_COMMAND);
+    prompt = get_prompt(info->prompttype);
     if (do_job_control)
 	print_job_status_all();
-    if (info->type == 1)
+    if (info->prompttype == 1)
 	check_mail();
     restore_parse_state(state);
     /* Note: no commands must be executed between `print_job_status_all' here
@@ -256,17 +207,17 @@ inputresult_T input_readline(struct xwcsbuf_T *buf, void *inputinfo)
 
 #if YASH_ENABLE_LINEEDIT
     /* read a line using line editing */
-    if (shopt_lineedit != shopt_nolineedit) {
+    if (info->fileinfo->fd == STDIN_FILENO
+	    && shopt_lineedit != shopt_nolineedit) {
 	wchar_t *line;
 	inputresult_T result;
 
-	unset_nonblocking(STDIN_FILENO);
 	result = le_readline(prompt, &line);
 	if (result != INPUT_ERROR) {
 	    free_prompt(prompt);
 	    if (result == INPUT_OK) {
-		if (info->type == 1)
-		    info->type = 2;
+		if (info->prompttype == 1)
+		    info->prompttype = 2;
 #if YASH_ENABLE_HISTORY
 		add_history(line);
 #endif
@@ -280,23 +231,20 @@ inputresult_T input_readline(struct xwcsbuf_T *buf, void *inputinfo)
     /* read a line without line editing */
     print_prompt(prompt.main);
     print_prompt(prompt.styler);
-    if (info->type == 1)
-	info->type = 2;
+    if (info->prompttype == 1)
+	info->prompttype = 2;
 
     int result;
 #if YASH_ENABLE_HISTORY
     size_t oldlen = buf->length;
 #endif
-    if (info->fp == stdin)
-	result = input_stdin(buf, NULL);
-    else
-	result = input_file(buf, info->fp);
+    result = input_file(buf, info->fileinfo);
 
     print_prompt(PROMPT_RESET);
     free_prompt(prompt);
 
 #if YASH_ENABLE_HISTORY
-    if (info->type == 2)
+    if (info->prompttype == 2)
 	add_history(buf->contents + oldlen);
 #endif
     return result;
@@ -493,21 +441,6 @@ done:
     wb_destroy(&buf);
 }
 
-/* Sets O_NONBLOCK flag of the specified file descriptor.
- * If `fd' is negative, does nothing.
- * Returns true if successful, or false otherwise, with `errno' set. */
-bool set_nonblocking(int fd)
-{
-    if (fd >= 0) {
-	int flags = fcntl(fd, F_GETFL);
-	if (flags < 0)
-	    return false;
-	if (!(flags & O_NONBLOCK))
-	    return fcntl(fd, F_SETFL, flags | O_NONBLOCK) != -1;
-    }
-    return true;
-}
-
 /* Unsets O_NONBLOCK flag of the specified file descriptor.
  * If `fd' is negative, does nothing.
  * Returns true if successful, or false otherwise, with `errno' set. */
