@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* plist.c: modifiable list of pointers */
-/* (C) 2007-2009 magicant */
+/* (C) 2007-2010 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -140,10 +140,10 @@ plist_T *pl_replace(
 
 /* Appends `p' to the end of pointer list `list' as a new element.
  * `p' may be NULL or a pointer to the list itself. */
-plist_T *pl_add(plist_T *list, void *p)
+plist_T *pl_add(plist_T *list, const void *p)
 {
     pl_ensuremax(list, list->length + 1);
-    list->contents[list->length++] = p;
+    list->contents[list->length++] = (void *) p;
     list->contents[list->length] = NULL;
     return list;
 }
