@@ -598,18 +598,26 @@ bool wait_for_input(int fd, bool trap, int timeout)
 /* Handles SIGCHLD if caught. */
 void handle_sigchld(void)
 {
+    static bool print_status = false;
+
     if (sigchld_received) {
         sigchld_received = false;
         do_wait();
+	print_status = true;
+    }
 
+    if (print_status) {
 	/* print job status if the notify option is set */
 #if YASH_ENABLE_LINEEDIT
-	if (le_state == LE_STATE_ACTIVE_COMPLETING)
-	    return;
-	if ((shopt_notify || shopt_notifyle) && le_state == LE_STATE_ACTIVE) {
-	    le_suspend_readline();
-	    print_job_status_all();
-	    le_resume_readline();
+	if (le_state & LE_STATE_ACTIVE) {
+	    if (!(le_state & LE_STATE_COMPLETING)) {
+		if (shopt_notify || shopt_notifyle) {
+		    le_suspend_readline();
+		    print_status = false;
+		    print_job_status_all();
+		    le_resume_readline();
+		}
+	    }
 	} else
 #endif
 	if (shopt_notify) {
@@ -618,6 +626,7 @@ void handle_sigchld(void)
 	    sigaddset(&ss, SIGTTOU);
 	    sigemptyset(&savess);
 	    sigprocmask(SIG_BLOCK, &ss, &savess);
+	    print_status = false;
 	    print_job_status_all();
 	    sigprocmask(SIG_SETMASK, &savess, NULL);
 	}
@@ -640,7 +649,7 @@ int handle_traps(void)
 #if YASH_ENABLE_LINEEDIT
     /* Don't handle traps during command line completion. Otherwise, the command
      * line would be messed up! */
-    if (le_state == LE_STATE_ACTIVE_COMPLETING)
+    if (le_state & LE_STATE_COMPLETING)
 	return 0;
 #endif
 
@@ -707,7 +716,7 @@ int handle_traps(void)
     if (state)
 	restore_parse_state(state);
 #if YASH_ENABLE_LINEEDIT
-    if (shopt_notifyle && le_state == LE_STATE_SUSPENDED)
+    if (shopt_notifyle && (le_state & LE_STATE_SUSPENDED))
 	print_job_status_all();
     le_resume_readline();
 #endif
@@ -913,7 +922,8 @@ void reset_sigint(void)
 
 #if YASH_ENABLE_LINEEDIT
 
-static void sig_new_candidate(const xfnmatch_T *pat, int num, xwcsbuf_T *name)
+static void sig_new_candidate(
+	const le_compopt_T *compopt, int num, xwcsbuf_T *name)
     __attribute__((nonnull));
 
 #ifdef SIGWINCH
@@ -936,26 +946,25 @@ void reset_sigwinch(void)
 #endif
 }
 
-/* Generates completion candidates for signal names that match the glob pattern
- * in the specified context. */
+/* Generates completion candidates for signal names matching the pattern. */
 /* The prototype of this function is declared in "lineedit/complete.h". */
-void generate_signal_candidates(le_candgentype_T type, le_context_T *context)
+void generate_signal_candidates(const le_compopt_T *compopt)
 {
-    if (!(type & CGT_SIGNAL))
+    if (!(compopt->type & CGT_SIGNAL))
 	return;
 
-    le_compdebug("adding signals for pattern \"%ls\"", context->pattern);
-    if (!le_compile_cpattern(context))
+    le_compdebug("adding signal name candidates");
+    if (!le_compile_cpatterns(compopt))
 	return;
 
-    bool prefix = matchwcsprefix(context->src, L"SIG");
+    bool prefix = matchwcsprefix(compopt->src, L"SIG");
     xwcsbuf_T buf;
     wb_init(&buf);
     for (const signal_T *s = signals; s->no; s++) {
 	if (prefix)
 	    wb_cat(&buf, L"SIG");
 	wb_cat(&buf, s->name);
-	sig_new_candidate(context->cpattern, s->no, &buf);
+	sig_new_candidate(compopt, s->no, &buf);
     }
 #if defined SIGRTMIN && defined SIGRTMAX
     int sigrtmin = SIGRTMIN, sigrtmax = SIGRTMAX;
@@ -965,25 +974,24 @@ void generate_signal_candidates(le_candgentype_T type, le_context_T *context)
 	wb_cat(&buf, L"RTMIN");
 	if (s != sigrtmin)
 	    wb_wprintf(&buf, L"+%d", s - sigrtmin);
-	sig_new_candidate(context->cpattern, s, &buf);
+	sig_new_candidate(compopt, s, &buf);
 
 	if (prefix)
 	    wb_cat(&buf, L"SIG");
 	wb_cat(&buf, L"RTMAX");
 	if (s != sigrtmax)
 	    wb_wprintf(&buf, L"-%d", sigrtmax - s);
-	sig_new_candidate(context->cpattern, s, &buf);
+	sig_new_candidate(compopt, s, &buf);
     }
 #endif
     wb_destroy(&buf);
 }
 
-/* If the specified pattern matches the specified name, adds a new completion
- * candidate with the specified name and the description for the specified
- * signal. */
-void sig_new_candidate(const xfnmatch_T *pat, int num, xwcsbuf_T *name)
+/* If the pattern in `compopt' matches the specified signal name, adds a new
+ * completion candidate with the name and the description for the signal. */
+void sig_new_candidate(const le_compopt_T *compopt, int num, xwcsbuf_T *name)
 {
-    if (xfnm_wmatch(pat, name->contents).start != (size_t) -1) {
+    if (le_wmatch_comppatterns(compopt, name->contents)) {
 	xwcsbuf_T desc;
 	wb_init(&desc);
 #if HAVE_STRSIGNAL
@@ -993,7 +1001,7 @@ void sig_new_candidate(const xfnmatch_T *pat, int num, xwcsbuf_T *name)
 	else
 #endif
 	    wb_wprintf(&desc, L"%d", num);
-	le_new_candidate(CT_SIG, wb_towcs(name), wb_towcs(&desc));
+	le_new_candidate(CT_SIG, wb_towcs(name), wb_towcs(&desc), compopt);
 	wb_init(name);
     } else {
 	wb_clear(name);
@@ -1016,9 +1024,9 @@ static void signal_job(int signum, const wchar_t *jobname)
 int trap_builtin(int argc, void **argv)
 {
     static const struct xoption long_options[] = {
-	{ L"print", xno_argument, L'p', },
+	{ L"print", OPTARG_NONE, L'p', },
 #if YASH_ENABLE_HELP
-	{ L"help",  xno_argument, L'-', },
+	{ L"help",  OPTARG_NONE, L'-', },
 #endif
 	{ NULL, 0, 0, },
     };
